/**********************************************************************
Copyright (c) 2009 Andy Jefferson and others. All rights reserved.
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Contributors:
   ...
**********************************************************************/
package org.datanucleus.jpa.criteria;

import java.util.ArrayList;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Set;

import javax.persistence.criteria.CriteriaQuery;
import javax.persistence.criteria.Expression;
import javax.persistence.criteria.Order;
import javax.persistence.criteria.ParameterExpression;
import javax.persistence.criteria.Predicate;
import javax.persistence.criteria.Root;
import javax.persistence.criteria.Selection;
import javax.persistence.criteria.Subquery;
import javax.persistence.metamodel.EntityType;
import javax.persistence.metamodel.Metamodel;

import org.datanucleus.jpa.metamodel.MetamodelImpl;

/**
 * Implementation of JPA2 Criteria "Query".
 */
public class CriteriaQueryImpl<T> implements CriteriaQuery<T>
{
    private MetamodelImpl model;

    private boolean distinct;
    private List<Selection<?>> selections;
    private Class<T> resultClass;
    private Set<RootImpl<?>> from;
    private PredicateImpl where;
    private List<Expression<?>> groups;
    private PredicateImpl having;
    private List<Order> orders;

    private List<Subquery<?>> subqueries;

//    private Map params;

    /**
     * Constructor for a criteria query for the supplied model and result class.
     * @param model The model
     * @param resultClass Result class (if any)
     */
    public CriteriaQueryImpl(MetamodelImpl model, Class<T> resultClass)
    {
        this.model = model;
        this.resultClass = resultClass;
    }

    /**
     * Accessor for the model used by this query.
     * @return The model
     */
    public Metamodel getMetamodel()
    {
        return model;
    }

    /* (non-Javadoc)
     * @see javax.persistence.criteria.CriteriaQuery#distinct(boolean)
     */
    public CriteriaQuery<T> distinct(boolean flag)
    {
        this.distinct = flag;
        return this;
    }

    /* (non-Javadoc)
     * @see javax.persistence.criteria.AbstractQuery#isDistinct()
     */
    public boolean isDistinct()
    {
        return this.distinct;
    }

    /* (non-Javadoc)
     * @see javax.persistence.criteria.CriteriaQuery#select(javax.persistence.criteria.Selection)
     */
    public CriteriaQuery<T> select(Selection<? extends T> select)
    {
        selections = new ArrayList<Selection<?>>();
        selections.add(select);
        return this;
    }

    /* (non-Javadoc)
     * @see javax.persistence.criteria.CriteriaQuery#multiselect(java.util.List)
     */
    public CriteriaQuery<T> multiselect(List<Selection<?>> selects)
    {
        selections = new ArrayList<Selection<?>>(selects);
        return this;
    }

    /* (non-Javadoc)
     * @see javax.persistence.criteria.CriteriaQuery#multiselect(javax.persistence.criteria.Selection<?>[])
     */
    public CriteriaQuery<T> multiselect(Selection<?>... selects)
    {
        selections = new ArrayList<Selection<?>>();
        for (int i=0;i<selects.length;i++)
        {
            selections.add(selects[i]);
        }
        return this;
    }

    /* (non-Javadoc)
     * @see javax.persistence.criteria.AbstractQuery#getSelection()
     */
    public Selection<T> getSelection()
    {
        if (selections != null)
        {
            return (Selection<T>) selections.get(0);
        }
        return null;
    }

    /* (non-Javadoc)
     * @see javax.persistence.criteria.AbstractQuery#getResultType()
     */
    public Class<T> getResultType()
    {
        return resultClass;
    }

    /* (non-Javadoc)
     * @see javax.persistence.criteria.AbstractQuery#from(java.lang.Class)
     */
    public <X> Root<X> from(Class<X> cls)
    {
        EntityType<X> entity = model.entity(cls);
        if (entity == null)
        {
            throw new IllegalArgumentException("The specified class (" + cls.getName() + ") is not an entity");
        }

        return from(entity);
    }

    /* (non-Javadoc)
     * @see javax.persistence.criteria.AbstractQuery#from(javax.persistence.metamodel.EntityType)
     */
    public <X> Root<X> from(EntityType<X> type)
    {
        if (from == null)
        {
            from = new HashSet<RootImpl<?>>();
        }
        RootImpl root = new RootImpl<X>(type);
        from.add(root);
        return root;
    }

    /* (non-Javadoc)
     * @see javax.persistence.criteria.AbstractQuery#getRoots()
     */
    public Set<Root<?>> getRoots()
    {
        if (from == null)
        {
            return null;
        }
        return new HashSet<Root<?>>(from);
    }

    /* (non-Javadoc)
     * @see javax.persistence.criteria.CriteriaQuery#where(javax.persistence.criteria.Expression)
     */
    public CriteriaQuery<T> where(Expression<Boolean> expr)
    {
        if (expr == null)
        {
            where = null;
            return this;
        }
        where = (PredicateImpl)expr;
        return this;
    }

    /* (non-Javadoc)
     * @see javax.persistence.criteria.CriteriaQuery#where(javax.persistence.criteria.Predicate[])
     */
    public CriteriaQuery<T> where(Predicate... exprs)
    {
        if (exprs == null)
        {
            where = null;
            return this;
        }
        // Set where clause using all of the predicates
        // TODO Do "and" of the expressions
//        where = new PredicateImpl.And(exprs);
        return this;
    }

    /* (non-Javadoc)
     * @see javax.persistence.criteria.AbstractQuery#getRestriction()
     */
    public Predicate getRestriction()
    {
        return where;
    }

    /* (non-Javadoc)
     * @see javax.persistence.criteria.CriteriaQuery#getParameters()
     */
    public Set<ParameterExpression<?>> getParameters()
    {
        // TODO Auto-generated method stub
        return null;
    }

    /* (non-Javadoc)
     * @see javax.persistence.criteria.CriteriaQuery#groupBy(javax.persistence.criteria.Expression<?>[])
     */
    public CriteriaQuery<T> groupBy(Expression<?>... exprs)
    {
        if (exprs == null)
        {
            groups = null;
            return this;
        }
        groups = new ArrayList<Expression<?>>();
        for (int i=0;i<exprs.length;i++)
        {
            groups.add(exprs[i]);
        }
        return this;
    }

    /* (non-Javadoc)
     * @see javax.persistence.criteria.CriteriaQuery#groupBy(java.util.List)
     */
    public CriteriaQuery<T> groupBy(List<Expression<?>> exprs)
    {
        if (exprs == null)
        {
            groups = null;
            return this;
        }
        groups = new ArrayList<Expression<?>>();
        groups.addAll(exprs);
        return this;
    }

    /* (non-Javadoc)
     * @see javax.persistence.criteria.AbstractQuery#getGroupList()
     */
    public List<Expression<?>> getGroupList()
    {
        if (groups == null)
        {
            return null;
        }
        return new ArrayList<Expression<?>>(groups);
    }

    /* (non-Javadoc)
     * @see javax.persistence.criteria.AbstractQuery#getGroupRestriction()
     */
    public Predicate getGroupRestriction()
    {
        return having;
    }

    /* (non-Javadoc)
     * @see javax.persistence.criteria.CriteriaQuery#having(javax.persistence.criteria.Expression)
     */
    public CriteriaQuery<T> having(Expression<Boolean> expr)
    {
        having = (PredicateImpl)expr;
        return this;
    }

    /* (non-Javadoc)
     * @see javax.persistence.criteria.CriteriaQuery#having(javax.persistence.criteria.Predicate[])
     */
    public CriteriaQuery<T> having(Predicate... exprs)
    {
        if (exprs == null)
        {
            having = null;
            return this;
        }

        // TODO Add this to any existing having
        return this;
    }

    /* (non-Javadoc)
     * @see javax.persistence.criteria.CriteriaQuery#getOrderList()
     */
    public List<Order> getOrderList()
    {
        if (orders == null)
        {
            return null;
        }
        return new ArrayList<Order>(orders);
    }

    /* (non-Javadoc)
     * @see javax.persistence.criteria.CriteriaQuery#orderBy(java.util.List)
     */
    public CriteriaQuery<T> orderBy(List<Order> list)
    {
        if (orders == null)
        {
            orders = null;
            return this;
        }

        orders = new ArrayList<Order>();
        orders.addAll(list);
        return this;
    }

    /* (non-Javadoc)
     * @see javax.persistence.criteria.CriteriaQuery#orderBy(javax.persistence.criteria.Order[])
     */
    public CriteriaQuery<T> orderBy(Order... arr)
    {
        if (orders == null)
        {
            orders = null;
            return this;
        }

        orders = new ArrayList<Order>();
        for (int i=0;i<arr.length;i++)
        {
            orders.add(arr[i]);
        }
        return this;
    }

    /* (non-Javadoc)
     * @see javax.persistence.criteria.AbstractQuery#subquery(java.lang.Class)
     */
    public <U> Subquery<U> subquery(Class<U> type)
    {
        if (subqueries == null)
        {
            subqueries = new ArrayList<Subquery<?>>();
        }
        Subquery<U> subquery = new SubqueryImpl<U>(type, this);
        subqueries.add(subquery);
        return subquery;
    }

    /**
     * Method to return a single-string representation of the criteria query in JPQL.
     * @return The single-string form
     */
    public String toString()
    {
        StringBuffer str = new StringBuffer();
        str.append("SELECT ");
        if (distinct)
        {
            str.append("DISTINCT ");
        }

        if (selections != null)
        {
            Iterator<Selection<?>> iter = selections.iterator();
            while (iter.hasNext())
            {
                Selection<?> select = iter.next();
                str.append(select.toString());
                if (iter.hasNext())
                {
                    str.append(",");
                }
            }
            str.append(" ");
        }

        // FROM clause
        str.append("FROM ");
        Iterator<RootImpl<?>> fromIter = from.iterator();
        while (fromIter.hasNext())
        {
            RootImpl<?> root = fromIter.next();
            str.append(root.toString(true));
            if (fromIter.hasNext())
            {
                str.append(",");
            }
        }
        str.append(" ");

        if (where != null)
        {
            // WHERE clause
            str.append("WHERE ").append(where.toString());
        }

        if (groups != null)
        {
            // GROUP BY clause
            str.append("GROUP BY ");
            Iterator<Expression<?>> iter = groups.iterator();
            while (iter.hasNext())
            {
                Expression<?> groupExpr = iter.next();
                str.append(groupExpr.toString());
                if (iter.hasNext())
                {
                    str.append(",");
                }
            }
            str.append(" ");
        }

        if (having != null)
        {
            // HAVING clause
            str.append("HAVING ");
            str.append(having.toString());
        }

        if (orders != null)
        {
            // ORDER BY clause
            str.append("ORDER BY ");
            Iterator<Order> iter = orders.iterator();
            {
                Order order = iter.next();
                str.append(order.toString());
                if (iter.hasNext())
                {
                    str.append(",");
                }
            }
            str.append(" ");
        }

        return str.toString().trim();
    }
}